/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <sstream>

#include <fcntl.h>
//#include <string.h>
#include <sys/stat.h>
#include "monitor.h"

#define FMT_DEFAULT "%h %l %u %t \\\"%r\\\" %>s %b"
#define REC_LEN     1024

using namespace std;
/**
 * Return file path
 */
string vhost::filepath()
{
   return this->fields[vhost::FILE_FIELD];
}
/**
 * Return file path in a form suitable for open
 */
const char* vhost::filestr()
{
   return this->filepath().c_str();
}
/**
 * Process row format to produce and array of needed input tokens. This is needed
 * to assign slots for the values in the host field vector.
 */
void vhost::getNeeds(topvhost *app)
{
   unsigned i;
   int ch;

   for(i=0;i<app->row.size();i++) {
      string token = app->row[i].format;
      if (token[0]=='%') {
         size_t lh  = token.find("{");
         string key = lh==string::npos? "" : token.substr(lh);
         switch(ch = app->row[i].field) {
            case 'B':   case 'C':   case 'U':   case 'b':   case 'e':   case 'h':
            case 'i':   case 'r':   case 's':   case 'u':   case 'v':
               key += ch;
               app->needs.push_back(key);
            }
         }
      }
}
/**
 * Format vhost content using a sprintf like specification using the place holders:
 *
 *  %H  = hour
 *  %M  = minute
 *  %S  = second
 *  %Y  = year
 *  %d  = day
 *  %h  = remote host     - right aligned and truncated/padded to width
 *  %m  = month
 *  %o  = count (hits)
 *  %r  = request string  - left aligned and truncated/padded to width
 *  %v  = domain name     - left aligned and always padded to namelength
 */
void vhost::paintHost(topvhost *app)
{
   struct tm *ts = localtime(&this->ldtm);
   int idx = vhost::OUTPUT_FIELDS;
   unsigned i,j;
   int ch,w;

   for(j=0;j<app->row.size();j++) {
      string token = app->row[j].format;
      switch(ch=app->row[j].field) {
         case ' ':
           app->wputs(token.c_str());
           break;
         case 'H':   app->wprint(token,ts->tm_hour);
            break;
         case 'M':   app->wprint(token,ts->tm_min);
           break;
         case 'S':   app->wprint(token,ts->tm_sec);
           break;
         case 'Y':   app->wprint(token,ts->tm_year);
           break;
         case 'd':   app->wprint(token,ts->tm_mday);
           break;
         case 'm':   app->wprint(token,ts->tm_mon+1);
            break;
         case 'o':   app->wprint(token,this->hits);
            break;
         default:    // Cehirsv
            string val     = this->fields[idx++];
            unsigned width = ch=='v'? app->namelen:0;
            if (width==0) {
               token.erase(0,1);
               stringstream ss(token);
               if ( (ss >> w).fail() )
                  width = app->maxx;
               else if (w > 0)
                  width = w;
               }
            for(i=0;i<val.length() && i < width;i++) {
               switch(ch = val[i]) {
                  case '\b':  case '\t':  ch = ' ';   break;
                  case '\r':  case '\n':  ch = 0;     break;
                  }
               if (ch==0) break;
               app->wput(ch);
               }
            while(i++<width)
               app->wput(' ');
            break;
         }
      }
}
/**
 * Parse apache log format string to build a vector of identitifers for the white
 * space separated log record components. The output is used to identifer parameter
 * position and type - since this application only recognizes a few types, modifiers
 * and literals such as quotes can be ignored.
 */
void vhost::parseApacheLogFormat(topvhost *app)
{
   char   ch;
   bool   quoted = false;
   bool   quotearg = false;
   string suffix="cinpPt";
   string token;
   string::iterator format;

   if (app->logfmt.length() == 0)
      app->logfmt = FMT_DEFAULT;
   format = app->logfmt.begin();
   do {
      ch = format==app->logfmt.end()? 0 : *format++;
      if (ch=='\\' && format!=app->logfmt.end()) ch=*format++;
      switch(ch) {
         case '"':
            quoted = !quoted;
            break;
         case '%':
            quotearg = quoted;
            token=ch;
            break;
         case 0: case ' ':
            if (!quoted) {
               if (token[0]=='%')
                  app->gets.push_back(token.substr(1));
               else app->gets.push_back("");
               token="";
               }
            break;
         case '>': case '<':
            if (!quoted) break;
         default:
            token += ch;
        }
      } while(ch !=0);
}
/**
 * Parse a log record. The 0th slot of the rec array is used as the parse space for
 * incremental parsing. Each new record is counted as a 'hit'
 */
void vhost::parseLogRecord(topvhost *app,char *s, int ct)
{
   char              ch;
   string            key;
   int               n;

   this->meter += ct;
   while(ct--) {
    switch(ch=*s++) {
      case '"':
         this->iq = !this->iq;
         break;
      case ' ':
        if (this->iq || this->fields[0][0]=='[') {
            this->fields[0] += ch;
            break;
            }
         else if (!this->fields[0].length())
            break;
      case '\n':
   done:
         n = app->xfer[this->argn];
         if (n>0)
            this->fields[n] = this->fields[0];
         if (ch=='\n') {
            this->argn = 0;
            this->hits++;
            }
         else this->argn++;
         this->fields[0] = "";
         break;
      case ']':
         this->fields[0] += ch;
         if (!this->iq && this->fields[0][0]=='[')
            goto done;
         break;
      default:
         this->fields[0] += ch;
      }
    }
}
/**
 * Clean the dirty bit
 */
bool vhost::setClean(topvhost *app)
{
   char  buf[REC_LEN];
   int   ct;

   int fh = open(this->filestr(),O_RDONLY);
   if (fh!=-1) {
       if (lseek(fh,this->fsz,SEEK_SET)!=-1) {
         while((ct = read(fh,buf,sizeof(buf)))>0)
           this->parseLogRecord(app,buf,ct);
         this->fsz = lseek(fh,0,SEEK_CUR);
         }
      close(fh);
      return true;
      }
   return false;
}
/**
 * Save info for later processing
 */
bool vhost::setDirty(bool init)
{
   struct stat attrib;

   if (stat(this->filestr(),&attrib)==0) {
      this->ldtm  = attrib.st_mtime;
      this->dirty = init;
      if (!init)
         this->fsz = attrib.st_size;
      return true;
      }
   return false;
}
/**
 * Verify record by showing record, domain
 */
void vhost::verify()
{
   char buf[256];
   struct tm *ts = localtime(&this->ldtm);
   this->setDirty(false);
   strftime(buf,sizeof(buf),"%Y-%02m-%02d %02H:%02M:%02S",ts);
   cout << this->fields[this->nidx] << endl;
   cout << this->filepath() << " " << buf << " " << this->fsz << endl << endl;
}
